#! /usr/bin/env python

'''Parallel shell based on SSS PM interface'''
__revision__ = '$Revision: 1.16 $'

import atexit
from getopt import getopt, GetoptError
from elementtree.ElementTree import Element, XML, SubElement, tostring
from os import environ, getuid
from pwd import getpwuid
from sss.ssslib import comm_lib, LookupError as SSSLookupError
from sss.server import EventReceiver
from sys import argv

class PMEvent(EventReceiver):
    '''PMEvent waits until process group has completed'''
    def __init__(self, pgrp, debug=0):
        self.__subscriptions__ = [('process-manager', '**', pgrp)]
        EventReceiver.__init__(self, debug=debug)

    def HandleEvent(self, xe, portinfo):
        '''Shutdown when process_end event arrives'''
        for event in xe:
            if event.find("msg").text == "process_end":
                self.shut = 1
        return Element('event-ok')
    
if __name__ == '__main__':
    try:
        (opts, args) = getopt(argv[1:], 'aw:g:f:d')
    except GetoptError, msg:
        print "Usage: dsh.py [-a] [-d] [-g group] [-f file] [-w node1,node2] cmd"
        print msg
        raise SystemExit, 1

    mode = None
    for opt in opts:
        if opt[0] == '-a':
            mode = 'all'
        elif opt[0] == '-g':
            mode = 'group'
            tag = opt[1]
        elif opt[0] == '-f':
            mode = 'node'
            nodes = [item.strip() for item in open(opt[1]).readlines()]
        elif opt[0] == '-w':
            mode = 'node'
            nodes = opt[1].split(',')
        elif opt[0] == '-d':
            pass
        else:
            print "semantic error"
            raise SystemExit, 1

    if not args:
        print "Usage: dsh.py [-a] [-d] [-g group] [-f file] [-w node1,node2] cmd"
        print "One of -a, -g, -f, or -w are required"
        raise SystemExit, 1

    cmd = args[0]

    args = args[1:]

    if not mode:
        print "Usage: dsh.py [-a] [-d] [-g group] [-f file] [-w node1,node2] cmd"
        print "One of -a, -g, -f, or -w are required"
        raise SystemExit, 1

    submitter = getpwuid(getuid())[0]
    path = environ['PATH']
    cwd = environ['PWD']

    comm = comm_lib(debug = ('-d' in argv))

    #if mode=group, query the build-system to find out the list of nodes
    if mode == 'group':
        query = Element('get-tag', node='*')
        bsh = comm.ClientInit('build-system')
        comm.SendMessage(bsh, tostring(query))
        answer = comm.RecvMessage(bsh)
        comm.ClientClose(bsh)
        data = XML(answer)
        # not finished here but I am waiting to test it for real on my machine, wanted to checkin 
        # put the hostlist into nodes
        nodes = [tag.find('node').text for tag in XML.findall('tag') if tag.find('node') != None]
    
    # at this point, regardless of mode, nodes is populated
    numprocs = str(len(nodes))

    pmcmd = Element("create-process-group")
    SubElement(pmcmd, "totalprocs")
    pmcmd.find("totalprocs").text = numprocs
    SubElement(pmcmd, "submitter")
    pmcmd.find("submitter").text = submitter
    ps = SubElement(pmcmd, "process-spec")
    SubElement(ps, "path")
    ps.find("path").text = path
    SubElement(ps, "cwd")
    ps.find("cwd").text = cwd
    SubElement(ps, "exec")
    ps.find("exec").text = cmd
    hs = SubElement(pmcmd,'host-spec')
    hs.text = ('\n').join(nodes)
    for arg in range(len(args)):
        ae = SubElement(ps, 'arg')
        SubElement(ae, "idx")
        ae.find("idx").text = str(arg)
        SubElement(ae, "value")
        ae.find("value").text = args[arg]
    try:
        pm = comm.ClientInit('process-manager')
    except SSSLookupError:
        print "Process Manager down"
        raise SystemExit, 1
        
    comm.SendMessage(pm, tostring(pmcmd))
    ack = comm.RecvMessage(pm)
    comm.ClientClose(pm)

    xml = XML(ack)
    if xml.tag == 'error':
        print ack
        raise SystemExit, 1
    pgid = xml.find('pgid').text
    debug = argv.count('-d')

    server = PMEvent(pgid, debug=debug)
    server.ServeForever()
    server.Shutdown()
    atexit._exithandlers = []
    del server
    # Now the process is done; we need to collect output
    pm = comm.ClientInit('process-manager')
    comm.SendMessage(pm, "<get-process-group-info><process-group><pgid>%s</pgid><output>*</output><error>*</error></process-group></get-process-group-info>" % pgid)
    resp = comm.RecvMessage(pm)
    comm.ClientClose(pm)
    rxml = XML(resp)

    output = ''
    err = ''

    try:
        output = "".join([node.text for node in rxml.findall('.//output')])
    except:
        pass

    try:
        err = rxml.findall('.//error').text
    except:
        pass

    if output:
        print output,
    if err:
        print err,
        raise SystemExit, 1
