#!/usr/bin/env python

from cPickle import dumps, loads
from elementtree.ElementTree import Element, SubElement, tostring
from sys import argv
from signal import signal, SIGINT
from select import select
from syslog import syslog, LOG_ERR
from telnetlib import Telnet
from threading import Thread
from os import fork, close
from time import strftime, localtime


from sss.server import Server, BufferedPipe
from sss.ssslib import comm_lib
from sss.restriction import DataSet, Data

def control_tel(dev,action,portlist):
    devt=Telnet(dev)
    devt.read_until("Enter Selection>")
    devt.write("1\r")
    devt.read_until(">")
    for port in portlist:
        devt.write("%s %s\r" % (action.encode(), str(port)))
        # write chokes on unicode
        devt.read_until(">")
        devt.write("y\r")
        devt.write("\r")
        devt.read_until(">")
    devt.write("menu\r")
    devt.read_until(">")
    devt.write("6\r")
    devt.read_very_eager()
    devt.close()

class chiba_hardware_infrastructure(Server):
    __component__='cluster-hardware'
    __implementation__='chiba-chi'
    __statefields__= ['chidata']
    __dispatch__= { 'get-node-hardware':'chidata.Get',
                    'del-node-hardware':'chidata.Del',
                    'set-node-hardware':'XSet',
                    'add-node-hardware':'chidata.Add',
                    'power-control':'XPower'}
    __validate__ = 1

    def __setup__(self):
        self.chidata = DataSet('node-hardware-set', 'node-hardware', Data, None, False)

    def XSet(self, xml, (peer,port)):
        self.changed_nodes = {}
        result = self.chidata.Get(xml,(peer,port), self.Set)
        self.evt.WriteMessage(dumps(self.changed_nodes.keys()))
        return result

    def Set(self, item, attrib):
        for x in attrib.keys():
            if item.attr[x] != attrib[x]:
                item.attr[x]=attrib[x]
                self.changed_nodes[item.attr['node']]=1

    def XPower(self, xml, (peer,port)):
        ports={}
        powerdata = self.chidata.Get(xml,(peer,port), None)
        action = xml.attrib['type']
        for x in powerdata:
            dev = x.attrib['powerdev']
            port = x.attrib['powerport']
            if ports.has_key(dev):
                ports[dev].append(port)
            else:
                ports[dev]=[port]

        self.powevt.WriteMessage(dumps((action, ports)))
        return powerdata
            
    def ChildSigHand(self, signum, frame):
        raise SystemExit, 0

    def __setupChildren__(self):
        # Setup Child Process to talk to the event manager
        self.evt = BufferedPipe()
        self.pipes.append(self.evt)
        children = []
        pid = fork()
        if pid == 0:
            signal(SIGINT, self.ChildSigHand)
            self.readsockets.append(self.evt.r)
            self.socketdispatch[self.evt.r]=self.ProcessChanges
            self.ServeForever()
        else:
            close(self.evt.r)
            children.append(pid)

        # Setup child process to talk to power controllers
        self.powevt = BufferedPipe()
        self.pipes.append(self.powevt)
        pid = fork()
        if pid == 0:
            signal(SIGINT, self.ChildSigHand)
            self.readsockets.append(self.powevt.r)
            self.socketdispatch[self.powevt.r] = self.ProcessPowCont
            self.ServeForever()
        else:
            close(self.powevt.r)
            children.append(pid)

        return children

    def ProcessChanges(self, pipe, inl=None):
        while 1:
            try:
                (infd, out, x) = select([self.evt.r], [], [], 5)
            except:
                pass
            if infd != []:
                self.evt.ReadMessages()

            while self.evt.buffer:
                changed_nodes = loads(self.evt.ReadMessage())
        
                # Send state changed message
                doc = Element('add-event')
                for node in changed_nodes:
                    SubElement(doc, 'event', component=self.__component__, 
                               msg = 'state-changed', data=node, 
                               time = strftime("%c", localtime()))
                if len(doc) > 0:
                    try:
                        c = comm_lib()
                        h = c.ClientInit('event-manager')
                        c.SendMessage(h, tostring(doc))
                        c.RecvMessage(h)
                        c.ClientClose(h)
                    except:
                        syslog(LOG_ERR, "Failed to send state-changed event")
        
                del doc
        return 0

    def ProcessPowCont(self, pipe, inl=None):
        while 1:
            try:
                (infd, out, x) = select([self.powevt.r], [], [], 5)
            except:
                pass
            if infd != []:
                self.powevt.ReadMessages()

            while self.powevt.buffer:
                (action, ports) = loads(self.powevt.ReadMessage())
        
                thread_list = []
                for dev in ports.keys():
                    tempthread = Thread(target=control_tel, args=(dev, action, ports[dev]))
                    tempthread.start()
                    thread_list.append(tempthread)
        
                # Block on completion of all threads, to serialize power-control calls
                for thread in thread_list:
                    thread.join()
        return 0

if __name__ == "__main__":
    from getopt import getopt, GetoptError
    try:
        (opts, args) = getopt(argv[1:], 'd', ['daemon='])
    except GetoptError, msg:
        print "%s\nUsage:\nchi.py [-d] [--daemon <pidfile>]" % (msg)
        raise SystemExit, 1
    daemon = [x[1] for x in opts if x[0] == '--daemon']
    debug = len([x for x in opts if x[0] == '-d'])
    if daemon:
        from sss.daemonize import daemonize
        daemonize(daemon[0])
    server = chiba_hardware_infrastructure(debug=debug)
    server.ServeForever()

